import unittest

import numpy as np

from pocomc.tools import compute_ess


class ESSTestCase(unittest.TestCase):
    def test_ess_single_particle(self):
        self.assertEqual(compute_ess(np.array([1.0])), 1.0)
        self.assertEqual(compute_ess(np.array([251.0])), 1.0)
        self.assertEqual(compute_ess(np.array([-421.0])), 1.0)
        self.assertEqual(compute_ess(np.array([-421.125251])), 1.0)
        self.assertEqual(compute_ess(np.array([0.0])), 1.0)


if __name__ == '__main__':
    unittest.main()



import unittest
import torch
from pocomc.flow import Flow

class FlowTestCase(unittest.TestCase):
    @staticmethod
    def make_data():
        # Make a dataset to use in tests
        torch.manual_seed(0)
        n_data = 100
        n_dim = 4
        x = torch.randn(size=(n_data, n_dim)) * 1.5
        return x

    @torch.no_grad()
    def test_forward(self):
        """
        Test that the forward pass works without raising an error
        """
        torch.manual_seed(0)

        data = self.make_data()
        flow = Flow(n_dim=data.shape[1])
        z, _ = flow.forward(data)

        self.assertFalse(torch.any(torch.isnan(z)))
        self.assertFalse(torch.any(torch.isinf(z)))
        self.assertEqual(data.shape, z.shape)
        self.assertEqual(data.dtype, z.dtype)

    @torch.no_grad()
    def test_inverse(self):
        # Test that the inverse pass works without raising an error
        torch.manual_seed(0)

        z = self.make_data()
        flow = Flow(n_dim=z.shape[1])
        x, _ = flow.inverse(z)

        self.assertFalse(torch.any(torch.isnan(x)))
        self.assertFalse(torch.any(torch.isinf(x)))
        self.assertEqual(x.shape, z.shape)
        self.assertEqual(x.dtype, z.dtype)

    @torch.no_grad()
    def test_logprob(self):
        # Test that logprob works without raising an error
        torch.manual_seed(0)

        x = self.make_data()
        flow = Flow(n_dim=x.shape[1])

        log_prob = flow.log_prob(x)

        self.assertFalse(torch.any(torch.isnan(log_prob)))
        self.assertFalse(torch.any(torch.isinf(log_prob)))
        self.assertEqual(log_prob.shape, (x.shape[0],))
        self.assertEqual(x.dtype, log_prob.dtype)

    @torch.no_grad()
    def test_sample(self):
        # Test that sample works without raising an error
        torch.manual_seed(0)

        x_tmp = self.make_data()
        flow = Flow(n_dim=x_tmp.shape[1])
        x, _ = flow.sample(x_tmp.shape[0])

        self.assertFalse(torch.any(torch.isnan(x)))
        self.assertFalse(torch.any(torch.isinf(x)))
        self.assertEqual(x.shape, x_tmp.shape)
        self.assertEqual(x.dtype, x_tmp.dtype)

    @torch.no_grad()
    def test_reconstruction(self):
        # Test that latent points are reconstructed to be close enough to data points
        torch.manual_seed(0)

        x = self.make_data()
        flow = Flow(n_dim=x.shape[1])
        z, _ = flow.forward(x)
        x_reconstructed, _ = flow.inverse(z)

        self.assertFalse(torch.any(torch.isnan(x_reconstructed)))
        self.assertFalse(torch.any(torch.isinf(x_reconstructed)))
        self.assertEqual(x_reconstructed.shape, x.shape)
        self.assertEqual(x_reconstructed.dtype, x.dtype)
        self.assertTrue(torch.allclose(x, x_reconstructed, atol=1e-5))

    @torch.no_grad()
    def test_logprob_float32(self):
        # Test logprob when input is torch.float
        torch.manual_seed(0)

        x = self.make_data()
        x = x.float()
        flow = Flow(n_dim=x.shape[1])
        log_prob = flow.log_prob(x)

        self.assertFalse(torch.any(torch.isnan(log_prob)))
        self.assertFalse(torch.any(torch.isinf(log_prob)))
        self.assertEqual(log_prob.shape, (x.shape[0],))
        self.assertEqual(x.dtype, log_prob.dtype)

    @torch.no_grad()
    def test_logprob_float64(self):
        # Test logprob when input is torch.double
        torch.manual_seed(0)

        x = self.make_data()
        x = x.double()
        flow = Flow(n_dim=x.shape[1])
        with self.assertWarns(UserWarning):
            log_prob = flow.log_prob(x)

        self.assertFalse(torch.any(torch.isnan(log_prob)))
        self.assertFalse(torch.any(torch.isinf(log_prob)))
        self.assertEqual(log_prob.shape, (x.shape[0],))
        self.assertEqual(log_prob.dtype, torch.float32)

    @torch.no_grad()
    def test_logprob_single_example(self):
        # Test logprob when input is a single data point
        torch.manual_seed(0)

        x = self.make_data()
        x = x[0].reshape(1, -1)
        flow = Flow(n_dim=x.shape[1])
        log_prob = flow.log_prob(x)

        self.assertFalse(torch.any(torch.isnan(log_prob)))
        self.assertFalse(torch.any(torch.isinf(log_prob)))
        self.assertEqual(log_prob.shape, (x.shape[0],))
        self.assertEqual(x.dtype, log_prob.dtype)

    def test_logprob_backward(self):
        # Test backpropagation on the negative log likelihood
        torch.manual_seed(0)

        x = self.make_data()
        flow = Flow(n_dim=x.shape[1])
        log_prob = flow.log_prob(x)
        nll = -torch.mean(log_prob)
        nll.backward()

        for param in flow.flow.parameters():
            if param.requires_grad:
                self.assertIsNotNone(param.grad)
            else:
                self.assertIsNone(param.grad)

    @torch.no_grad()
    def test_logprob_inverse(self):
        # Test that the inverse logprob is the negative of the forward logprob
        torch.manual_seed(0)

        x = self.make_data()

        flow = Flow(n_dim=x.shape[1])
        z, logprob_forward = flow.forward(x)
        _, logprob_inverse = flow.inverse(z)

        #self.assertTrue(torch.allclose(logprob_forward, -logprob_inverse))
        torch.testing.assert_close(logprob_forward, -logprob_inverse)
        self.assertEqual(logprob_forward.shape, logprob_inverse.shape)
        self.assertEqual(logprob_forward.dtype, logprob_inverse.dtype)

    def test_fit(self):
        # Test that fit works without errors and check some basic functions afterwards
        torch.manual_seed(0)

        x = self.make_data()
        flow = Flow(n_dim=x.shape[1])
        flow.fit(x, epochs=5)

        z, _ = flow.forward(x)
        log_prob = flow.log_prob(x)
        x_samples, _ = flow.sample(x.shape[0])

        self.assertFalse(torch.any(torch.isnan(log_prob)))
        self.assertFalse(torch.any(torch.isinf(log_prob)))
        self.assertEqual(log_prob.shape, (x.shape[0],))
        self.assertEqual(x.dtype, log_prob.dtype)

        self.assertFalse(torch.any(torch.isnan(z)))
        self.assertFalse(torch.any(torch.isinf(z)))
        self.assertEqual(x.shape, z.shape)
        self.assertEqual(x.dtype, z.dtype)

        self.assertFalse(torch.any(torch.isnan(x_samples)))
        self.assertFalse(torch.any(torch.isinf(x_samples)))
        self.assertEqual(x.shape, x_samples.shape)
        self.assertEqual(x.dtype, x_samples.dtype)

    def test_logj(self):
        torch.manual_seed(0)

        x = self.make_data()
        flow = Flow(n_dim=x.shape[1])

        z, logj_forward = flow.forward(x)
        _, logj_inverse = flow.inverse(z)

        #assert torch.allclose(logj_forward, -logj_inverse, rtol=1e-4)
        torch.testing.assert_close(logj_forward, -logj_inverse)


if __name__ == '__main__':
    unittest.main()



__bibtex__ = """
@article{karamanis2022accelerating,
  title={Accelerating astronomical and cosmological inference with preconditioned Monte Carlo},
  author={Karamanis, Minas and Beutler, Florian and Peacock, John A and Nabergoj, David and Seljak, Uro{\v{s}}},
  journal={Monthly Notices of the Royal Astronomical Society},
  volume={516},
  number={2},
  pages={1644--1653},
  year={2022},
  publisher={Oxford University Press}
}

@article{karamanis2022pocomc,
  title={pocoMC: A Python package for accelerated Bayesian inference in astronomy and cosmology},
  author={Karamanis, Minas and Nabergoj, David and Beutler, Florian and Peacock, John A and Seljak, Uros},
  journal={arXiv preprint arXiv:2207.05660},
  year={2022}
}
"""
__url__ = "https://pocomc.readthedocs.io"
__author__ = "Minas Karamanis"
__email__ = "minaskar@gmail.com"
__license__ = "GPL-3.0"
__description__ = "A Python implementation of Preconditioned Monte Carlo for accelerated Bayesian Computation"


from .flow import *
from .sampler import *
from .prior import *
from ._version import version

__version__ = version



import numpy as np


def assert_array_2d(x: np.ndarray):
    if len(x.shape) != 2:
        raise ValueError(f"Input should have 2 dimensions, but got {len(x.shape)}")


def assert_array_1d(x: np.ndarray):
    if len(x.shape) != 1:
        raise ValueError(f"Input should have 1 dimension, but got {len(x.shape)}")


def assert_equal_type(x, y):
    if type(x) != type(y):
        raise ValueError(f"Expected inputs to have equal types, but got {type(x)} and {type(y)}")


def assert_arrays_equal_shape(x: np.ndarray,
                              y: np.ndarray):
    if x.shape != y.shape:
        raise ValueError(f"Inputs should have equal shape, but got {x.shape} and {y.shape}")


def assert_array_within_interval(x: np.ndarray,
                                 left: np.ndarray,
                                 right: np.ndarray,
                                 left_open: bool = False,
                                 right_open: bool = False):
    left = left.copy()
    left[np.isnan(left)] = -np.inf

    right = right.copy()
    right[np.isnan(right)] = np.inf

    if left_open and right_open:
        condition = (left < x) & (x < right)
        interval_string = f'({left}, {right})'
    elif left_open and not right_open:
        condition = (left < x) & (x <= right)
        interval_string = f'({left}, {right}]'
    elif not left_open and right_open:
        condition = (left <= x) & (x < right)
        interval_string = f'[{left}, {right})'
    else:
        condition = (left <= x) & (x <= right)
        interval_string = f'[{left}, {right}]'

    if not np.all(condition):
        x_min = np.min(x)
        x_max = np.max(x)
        raise ValueError(f"Expected input to be within interval {interval_string}, "
                         f"but got minimum = {x_min} and maximum = {x_max}")


def assert_array_float(x: np.ndarray):
    if not np.issubdtype(x.dtype, np.floating):
        raise ValueError(f"Expected input to have dtype float, but got {x.dtype}")



import unittest
import numpy as np

from scipy.stats import norm

from pocomc.sampler import Sampler
from pocomc.prior import Prior

class SamplerTestCase(unittest.TestCase):
    @staticmethod
    def log_likelihood_single(x):
        return np.sum(-0.5 * np.log(2 * np.pi) - 0.5 * x ** 2)

    @staticmethod
    def log_likelihood_vectorized(x):
        # Gaussian log likelihood with mu = 0, sigma = 1
        return np.sum(-0.5 * np.log(2 * np.pi) - 0.5 * x ** 2, axis=1)

    def test_run(self):
        
        n_dim = 2
        prior = Prior(n_dim*[norm(0, 1)])

        sampler = Sampler(
            prior=prior,
            likelihood=self.log_likelihood_single,
            train_config={'epochs': 1},
            random_state=0,
        )
        sampler.run()

    def test_run2(self):
        
        n_dim = 2
        prior = Prior(n_dim*[norm(0, 1)])

        sampler = Sampler(
            prior=prior,
            likelihood=self.log_likelihood_vectorized,
            vectorize=True,
            train_config={'epochs': 1},
            random_state=0,
        )
        sampler.run()


if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path
import numpy as np
from pocomc import Sampler
from pocomc import Prior

from scipy.stats import norm


class SamplerStateTestCase(unittest.TestCase):
    @staticmethod
    def log_likelihood_vectorized(x):
        # Gaussian log likelihood with mu = 0, sigma = 1
        return np.sum(-0.5 * np.log(2 * np.pi) - 0.5 * x ** 2, axis=1)

    def test_save(self):
        # Save PMC state.
        prior = Prior([norm(0, 1), norm(0, 1)])
        s = Sampler(prior, self.log_likelihood_vectorized, vectorize=True, train_config=dict(epochs=10), random_state=0)
        path = Path('pmc.state')
        s.save_state(path)
        self.assertTrue(path.exists())
        path.unlink()
        self.assertFalse(path.exists())

    def test_load(self):
        # Load PMC state.
        prior = Prior([norm(0, 1), norm(0, 1)])
        s = Sampler(prior, self.log_likelihood_vectorized, vectorize=True, train_config=dict(epochs=10), random_state=0)
        path = Path('pmc.state')
        s.save_state(path)
        self.assertTrue(path.exists())
        s.load_state(path)
        path.unlink()
        self.assertFalse(path.exists())

    def test_resume(self):
        # Run PMC. Then, pick an intermediate state and resume from that state.
        np.random.seed(0)
        prior = Prior([norm(0, 1), norm(0, 1)])
        s = Sampler(prior, self.log_likelihood_vectorized, vectorize=True, train_config=dict(epochs=10), random_state=0)
        s.run(save_every=1)  # Save every iteration

        # At this point, we would look at the directory and choose the file we want to load. In this example, we select
        # "pmc_1.state". Now we rerun the sampler starting from this path. We will not get the exact same
        # results due to RNG.

        self.assertTrue(Path("states/pmc_1.state").exists())
        self.assertTrue(Path("states/pmc_2.state").exists())
        self.assertTrue(Path("states/pmc_3.state").exists())

        s = Sampler(prior, self.log_likelihood_vectorized, vectorize=True, train_config=dict(epochs=10), random_state=0)
        s.run(resume_state_path="states/pmc_1.state")

        # Remove the generated state files
        #Path("states/pmc_1.state").unlink()
        #Path("states/pmc_2.state").unlink()
        #Path("states/pmc_3.state").unlink()
        p = Path("states").glob('**/*')
        files = [x for x in p if x.is_file()]
        for f in files:
            f.unlink()
        Path("states").rmdir()


if __name__ == '__main__':
    unittest.main()



import numpy as np

class Prior:
    """
    A class for priors.

    Parameters
    ----------
    dists : list of scipy.stats distributions
        A list of distributions for each parameter. The length of the list
        determines the dimension of the prior.

    Attributes
    ----------
    dists : list of scipy.stats distributions
        A list of distributions for each parameter. The length of the list
        determines the dimension of the prior.
    bounds : ndarray
        An array of shape (dim, 2) containing the lower and upper bounds for
        each parameter.
    dim : int
        The dimension of the prior.
    
    Methods
    -------
    logpdf(x)
        Returns the log of the probability density function evaluated at x.
    rvs(size=1)
        Returns a random sample from the prior.
    
    Examples
    --------
    >>> import numpy as np
    >>> from scipy.stats import norm, uniform
    >>> from pocomc.prior import Prior
    >>> dists = [norm(loc=0, scale=1), uniform(loc=0, scale=1)]
    >>> prior = Prior(dists)
    >>> prior.logpdf(np.array([0, 0]))
    -1.8378770664093453
    >>> prior.rvs()
    array([0.417022  , 0.72032449])
    >>> prior.bounds
    array([[-inf,  inf],
           [ 0. ,  1. ]])
    >>> prior.dim
    2

    Notes
    -----
    The logpdf method is implemented as a sum of the logpdf methods of the
    individual distributions. This is equivalent to assuming that the
    parameters are independent.

    The rvs method is implemented by sampling from each distribution
    independently and then transposing the result. This is equivalent to
    assuming that the parameters are independent.

    The bounds property is implemented by calling the support method of each
    distribution. This is equivalent to assuming that the parameters are
    independent.

    The dim property is implemented by returning the length of the dists
    attribute. This is equivalent to assuming that the parameters are
    independent.
    """

    def __init__(self, dists=None):
        self.dists = dists

    def logpdf(self, x):
        """
        Returns the log of the probability density function evaluated at x.

        Parameters
        ----------
        x : ndarray
            An array of shape (n, dim) containing n samples of the parameters.
        
        Returns
        -------
        logp : ndarray
            An array of shape (n,) containing the log of the probability
            density function evaluated at each sample.
        
        Examples
        --------
        >>> import numpy as np
        >>> from scipy.stats import norm, uniform
        >>> from pocomc.prior import Prior
        >>> dists = [norm(loc=0, scale=1), uniform(loc=0, scale=1)]
        >>> prior = Prior(dists)
        >>> prior.logpdf(np.array([0, 0]))
        -1.8378770664093453
        >>> prior.logpdf(np.array([[0, 0], [0, 0]]))
        array([-1.83787707, -1.83787707])
        """
        logp = np.zeros(len(x))
        for i, dist in enumerate(self.dists): 
            logp += dist.logpdf(x[:,i])
        return logp
    
    def rvs(self, size=1):
        """
        Returns a random sample from the prior.

        Parameters
        ----------
        size : int, optional
            The number of samples to return. The default is 1.
        
        Returns
        -------
        samples : ndarray
            An array of shape (size, dim) containing the samples.
    
        Examples
        --------
        >>> import numpy as np
        >>> from scipy.stats import norm, uniform
        >>> from pocomc.prior import Prior
        >>> dists = [norm(loc=0, scale=1), uniform(loc=0, scale=1)]
        >>> prior = Prior(dists)
        >>> prior.rvs()
        array([0.417022  , 0.72032449])
        >>> prior.rvs(size=2)
        array([[0.417022  , 0.72032449],
               [0.00011438, 0.30233257]])
        """
        samples = []
        for dist in self.dists:
            samples.append(dist.rvs(size=size))
        return np.transpose(samples)
    
    @property
    def bounds(self):
        """
        An array of shape (dim, 2) containing the lower and upper bounds for
        each parameter.

        Examples
        --------
        >>> import numpy as np
        >>> from scipy.stats import norm, uniform
        >>> from pocomc.prior import Prior
        >>> dists = [norm(loc=0, scale=1), uniform(loc=0, scale=1)]
        >>> prior = Prior(dists)
        >>> prior.bounds
        array([[-inf,  inf],
               [ 0. ,  1. ]])
        """
        bounds = []
        for dist in self.dists:
            bounds.append(dist.support())
        return np.array(bounds)
    
    @property
    def dim(self):
        """
        The dimension of the prior.

        Examples
        --------
        >>> import numpy as np
        >>> from scipy.stats import norm, uniform
        >>> from pocomc.prior import Prior
        >>> dists = [norm(loc=0, scale=1), uniform(loc=0, scale=1)]
        >>> prior = Prior(dists)
        >>> prior.dim
        2
        """
        return len(self.dists)


from typing import Any
import numpy as np

class Particles:
    """
    Class to store the particles and their associated weights.

    Parameters
    ----------
    n_particles : int
        Number of particles.
    n_dim : int
        Dimension of the parameter space.
    ess_threshold : float, optional
        Threshold for the effective sample size. If the effective sample size
        is below this threshold, the weights are set to zero. This is useful
        for the case where the effective sample size is very small, but not
        exactly zero, due to numerical errors.
    
    Attributes
    ----------
    n_particles : int
        Number of particles.
    n_dim : int
        Dimension of the parameter space.
    ess_threshold : float, optional
        Threshold for the effective sample size. If the effective sample size
        is below this threshold, the weights are set to zero. This is useful
        for the case where the effective sample size is very small, but not
        exactly zero, due to numerical errors.
    u : numpy.ndarray
        Array of shape (n_particles, n_dim) containing the particles.
    logdetj : numpy.ndarray
        Array of shape (n_particles,) containing the log-determinant of the
        Jacobian of the transformation from the unit hypercube to the
        parameter space.
    logl : numpy.ndarray
        Array of shape (n_particles,) containing the log-likelihoods.
    logp : numpy.ndarray
        Array of shape (n_particles,) containing the log-priors.
    logw : numpy.ndarray
        Array of shape (n_particles,) containing the log-weights.
    iter : numpy.ndarray
        Array of shape (n_particles,) containing the iteration number of each
        particle.
    logz : numpy.ndarray
        Array of shape (n_particles,) containing the log-evidence of each
        particle.
    calls : numpy.ndarray
        Array of shape (n_particles,) containing the number of likelihood
        evaluations of each particle.
    steps : numpy.ndarray
        Array of shape (n_particles,) containing the number of steps of each
        particle.
    efficiency : numpy.ndarray
        Array of shape (n_particles,) containing the efficiency of each
        particle.
    ess : numpy.ndarray
        Array of shape (n_particles,) containing the effective sample size of
        each particle.
    accept : numpy.ndarray
        Array of shape (n_particles,) containing the acceptance rate of each
        particle.
    beta : numpy.ndarray
        Array of shape (n_particles,) containing the inverse temperature of
        each particle.
    """

    def __init__(self, n_particles, n_dim, ess_threshold=None):
        self.n_particles = n_particles
        self.n_dim = n_dim
        self.ess_threshold = ess_threshold

        self.past = dict(
            u = [],
            x = [],
            logdetj = [],
            logl = [],
            logp = [],
            logw = [],
            iter = [],
            logz = [],
            calls = [],
            steps = [],
            efficiency = [],
            ess = [],
            accept = [],
            beta = [],
        )

        self.results_dict = None

    def update(self, data):
        """
        Update the particles with the given data.

        Parameters
        ----------
        data : dict
            Dictionary containing the data to be added to the particles.
        
        Notes
        -----
        The dictionary must contain the following keys:
            u : numpy.ndarray
                Array of shape (n_particles, n_dim) containing the particles.
            logdetj : numpy.ndarray
                Array of shape (n_particles,) containing the log-determinant
                of the Jacobian of the transformation from the unit hypercube
                to the parameter space.
            logl : numpy.ndarray
                Array of shape (n_particles,) containing the log-likelihoods.
            logp : numpy.ndarray
                Array of shape (n_particles,) containing the log-priors.
            logw : numpy.ndarray
                Array of shape (n_particles,) containing the log-weights.
            iter : numpy.ndarray
                Array of shape (n_particles,) containing the iteration number
                of each particle.
            logz : numpy.ndarray
                Array of shape (n_particles,) containing the log-evidence of
                each particle.
            calls : numpy.ndarray
                Array of shape (n_particles,) containing the number of
                likelihood evaluations of each particle.
            steps : numpy.ndarray
                Array of shape (n_particles,) containing the number of steps
                of each particle.
            efficiency : numpy.ndarray
                Array of shape (n_particles,) containing the efficiency of
                each particle.
            ess : numpy.ndarray
                Array of shape (n_particles,) containing the effective sample
                size of each particle.
            accept : numpy.ndarray
                Array of shape (n_particles,) containing the acceptance rate
                of each particle.
            beta : numpy.ndarray
                Array of shape (n_particles,) containing the inverse
                temperature of each particle.
        """
        for key in data.keys():
            if key in self.past.keys():
                value = data.get(key)
                # Save to past states
                self.past.get(key).append(value)

    def pop(self, key):
        """
        Remove the last element of the given key.

        Parameters
        ----------
        key : str
            Key of the element to be removed.
        
        Notes
        -----
        This method is useful to remove the last element of the particles
        after the resampling step.
        """
        _ = self.past.get(key).pop()

    def get(self, key, index=None, flat=False):
        """
        Get the element of the given key.

        Parameters
        ----------
        key : str
            Key of the element to be returned.
        index : int, optional
            Index of the element to be returned. If None, all elements are
            returned.
        flat : bool, optional
            If True, the elements are returned as a flattened array. Otherwise,
            the elements are returned as a numpy.ndarray.
        
        Returns
        -------
        element : numpy.ndarray
            Array of shape (n_particles,) or (n_particles, n_dim) containing
            the elements of the given key.
        
        Notes
        -----
        If index is None, the elements are returned as a numpy.ndarray. If
        index is not None, the elements are returned as a numpy.ndarray with
        shape (n_dim,). If flat is True, the elements are returned as a
        flattened array.

        Examples
        --------
        >>> particles = Particles(n_particles=10, n_dim=2)
        >>> particles.update(dict(u=np.random.randn(10,2)))
        >>> particles.get("u").shape
        (10, 2)
        >>> particles.get("u", index=0).shape
        (2,)
        >>> particles.get("u", index=0, flat=True).shape
        (2,)
        >>> particles.get("u", index=None, flat=True).shape
        (20,)
        """
        if index is None:
            if flat:
                return np.concatenate(self.past.get(key))
            else:
                return np.asarray(self.past.get(key))
        else:
            return self.past.get(key)[index]
        
    def compute_logw(self, beta, ess_threshold=None):
        """
        Compute the log-weights of the particles for the given inverse
        temperature.

        Parameters
        ----------
        beta : float
            Inverse temperature.
        ess_threshold : float, optional
            Threshold for the effective sample size. If the effective sample
            size is below this threshold, the weights are set to zero. This is
            useful for the case where the effective sample size is very small,
            but not exactly zero, due to numerical errors.
        
        Returns
        -------
        logw : numpy.ndarray
            Array of shape (n_particles,) containing the log-weights.
        
        Notes
        -----
        The log-weights are computed as
            logw = (beta - beta_original) * logl
        where beta_original is the inverse temperature of each particle and
        logl is the log-likelihood of each particle.

        Examples
        --------
        >>> particles = Particles(n_particles=10, n_dim=2)
        >>> particles.update(dict(u=np.random.randn(10,2)))
        >>> particles.compute_logw(beta=1.0).shape
        (10,)
        """
        logl = self.get("logl")
        beta_original = self.get("beta")

        logw = (beta - beta_original[:,None]) * logl
        logw -= np.logaddexp.reduce(logw, axis=1)[:,None]
        log_ess = - np.logaddexp.reduce(2.0 * logw, axis=1)
        ###
        if ess_threshold is not None:
            mask = np.exp(log_ess) < ess_threshold
            log_ess[mask] = -1e300
        ###
        log_ess -= np.logaddexp.reduce(log_ess)
        logw += log_ess[:,None]
        logw = logw.reshape(-1)
        logw -= np.logaddexp.reduce(logw)
        
        return logw
    
    def compute_logz(self, beta):
        """
        Compute the log-evidence of the particles for the given inverse
        temperature.

        Parameters
        ----------
        beta : float
            Inverse temperature.
        
        Returns
        -------
        logz : float
            Log-evidence.

        Notes
        -----
        The log-evidence is computed as
            logz = logsumexp(log_ess + logz + logz_increments)
        where log_ess is the log-effective sample size of each particle,
        logz is the log-evidence of each particle and logz_increments is the
        log-evidence increment of each particle.

        Examples
        --------
        >>> particles = Particles(n_particles=10, n_dim=2)
        >>> particles.update(dict(u=np.random.randn(10,2)))
        >>> particles.compute_logz(beta=1.0)
        """
        logz = self.get("logz")

        logl = self.get("logl")
        beta_original = self.get("beta")

        logw = (beta - beta_original[:,None]) * logl
        logw_normed = logw - np.logaddexp.reduce(logw, axis=1)[:,None]

        logz_increments = np.logaddexp.reduce(logw, axis=1) - np.log(logw.shape[1])
        
        log_ess = - np.logaddexp.reduce(2.0 * logw_normed, axis=1)
        log_ess_normed = log_ess - np.logaddexp.reduce(log_ess)

        logz_new = np.logaddexp.reduce(log_ess_normed + logz + logz_increments)

        return logz_new
    
    def compute_results(self):
        """
        Compute the results of the particles.

        Returns
        -------
        results_dict : dict
            Dictionary containing the results of the particles.

        Notes
        -----
        The dictionary contains the following keys:
            u : numpy.ndarray
                Array of shape (n_particles, n_dim) containing the particles.
            logdetj : numpy.ndarray
                Array of shape (n_particles,) containing the log-determinant
                of the Jacobian of the transformation from the unit hypercube
                to the parameter space.
            logl : numpy.ndarray
                Array of shape (n_particles,) containing the log-likelihoods.
            logp : numpy.ndarray
                Array of shape (n_particles,) containing the log-priors.
            logw : numpy.ndarray
                Array of shape (n_particles,) containing the log-weights.
            iter : numpy.ndarray
                Array of shape (n_particles,) containing the iteration number
                of each particle.
            logz : numpy.ndarray
                Array of shape (n_particles,) containing the log-evidence of
                each particle.
            calls : numpy.ndarray
                Array of shape (n_particles,) containing the number of
                likelihood evaluations of each particle.
            steps : numpy.ndarray
                Array of shape (n_particles,) containing the number of steps
                of each particle.
            efficiency : numpy.ndarray
                Array of shape (n_particles,) containing the efficiency of
                each particle.
            ess : numpy.ndarray
                Array of shape (n_particles,) containing the effective sample
                size of each particle.
            accept : numpy.ndarray
                Array of shape (n_particles,) containing the acceptance rate
                of each particle.
            beta : numpy.ndarray
                Array of shape (n_particles,) containing the inverse
                temperature of each particle.

        Examples
        --------
        >>> particles = Particles(n_particles=10, n_dim=2)
        >>> particles.update(dict(u=np.random.randn(10,2)))
        >>> particles.compute_results().keys()
        dict_keys(['u', 'logdetj', 'logl', 'logp', 'logw', 'iter', 'logz', 'calls', 'steps', 'efficiency', 'ess', 'accept', 'beta'])
        """
        if self.results_dict is None:
            self.results_dict = dict()
            for key in self.past.keys():
                self.results_dict[key] = self.get(key)
        
            logl = self.get("logl")
            beta_original = self.get("beta")

            logw = (1.0 - beta_original[:,None]) * logl
            logw -= np.logaddexp.reduce(logw, axis=1)[:,None]
            log_ess = - np.logaddexp.reduce(2.0 * logw, axis=1)
            ###
            if self.ess_threshold is not None:
                mask = np.exp(log_ess) < self.ess_threshold
                log_ess[mask] = -1e300
            ###
            log_ess -= np.logaddexp.reduce(log_ess)
            logw += log_ess[:,None]

            self.results_dict["logw"] = logw
            self.results_dict["ess"] = np.exp(log_ess)

        return self.results_dict




from pathlib import Path
from typing import Union

import dill
import numpy as np
import torch

from .mcmc import preconditioned_pcn, preconditioned_rwm, pcn, rwm
from .tools import systematic_resample, FunctionWrapper, numpy_to_torch, torch_to_numpy, trim_weights, ProgressBar, flow_numpy_wrapper
from .scaler import Reparameterize
from .flow import Flow
from .particles import Particles
from .geometry import Geometry

class Sampler:
    r"""Preconditioned Monte Carlo class.

    Parameters
    ----------
    prior : callable
        Class implementing the prior distribution.
    likelihood : callable
        Function returning the log likelihood of a set of parameters.
    n_dim : int
        The total number of parameters/dimensions (Optional as it can be infered from the prior class).
    n_ess : int
        The effective sample size maintained during the run (default is ``n_ess=1000``).
    n_active : int
        The number of active particles (default is ``n_active=250``). It must be smaller than ``n_ess``.
    likelihood_args : list
        Extra arguments to be passed to likelihood
        (default is ``likelihood_args=None``).
    likelihood_kwargs : dict
        Extra arguments to be passed to likelihood
        (default is ``likelihood_kwargs=None``).
    vectorize : bool
        If True, vectorize ``likelihood``
        calculation (default is ``vectorize=False``).
    pool : pool
        Provided ``MPI`` or ``multiprocessing`` pool for
        parallelisation (default is ``pool=None``).
    flow : ``torch.nn.Module`` or ``None``
        Normalizing flow (default is ``None``). The default is a Masked Autoregressive Flow
        (MAF) with 6 blocks of 3x64 layers and residual connections.
    train_config : dict or ``None``
        Configuration for training the normalizing flow
        (default is ``train_config=None``). Options include a dictionary with the following
        keys: ``"validation_split"``, ``"epochs"``, ``"batch_size"``, ``"patience"``,
        ``"learning_rate"``, ``"annealing"``, ``"gaussian_scale"``, ``"laplace_scale"``,
        ``"noise"``, ``"shuffle"``, ``"clip_grad_norm"``, ``"verbose"``.
    precondition : bool
        If True, use preconditioned MCMC (default is ``precondition=True``). If False,
        use standard MCMC without normalizing flow.
    n_prior : int
        Number of prior samples to draw (default is ``n_prior=2*(n_ess//n_active)*n_active``).
    sample : ``str``
        Type of MCMC sampler to use (default is ``sample="pcn"``). Options are
        ``"pcn"`` (Preconditioned Crank-Nicolson) or ``"rwm"`` (Random-Walk Metropolis).
    max_steps : int
        Maximum number of MCMC steps (default is ``max_steps=5*n_dim``).
    patience : int
        Number of steps for early stopping of MCMC (default is ``patience=None``). If ``patience=None``,
        MCMC terminates automatically.
    ess_threshold : int
        Effective sample size threshold for resampling (default is ``ess_threshold=4*n_dim``).
    output_dir : ``str`` or ``None``
        Output directory for storing the state files of the
        sampler. Default is ``None`` which creates a ``states``
        directory. Output files can be used to resume a run.
    output_label : ``str`` or ``None``
        Label used in state files. Defaullt is ``None`` which
        corresponds to ``"pmc"``. The saved states are named
        as ``"{output_dir}/{output_label}_{i}.state"`` where
        ``i`` is the iteration index.  Output files can be
        used to resume a run.
    random_state : int or ``None``
        Initial random seed.
    """

    def __init__(self,
                 prior: callable,
                 likelihood: callable,
                 n_dim: int = None,
                 n_ess: int = 1000,
                 n_active: int = 250,
                 likelihood_args: list = None,
                 likelihood_kwargs: dict = None,
                 vectorize: bool = False,
                 pool=None,
                 flow=None,
                 train_config: dict = None,
                 precondition: bool = True,
                 n_prior: int = None,
                 sample: str = None,
                 max_steps: int = None,
                 patience: int = None,
                 ess_threshold: int = None,
                 output_dir: str = None,
                 output_label: str = None,
                 random_state: int = None):

        # Random seed
        if random_state is not None:
            np.random.seed(random_state)
            torch.manual_seed(random_state)
        self.random_state = random_state

        # Prior
        self.prior = prior
        self.log_prior = self.prior.logpdf
        self.sample_prior = self.prior.rvs
        self.bounds = self.prior.bounds

        # Log likelihood
        self.log_likelihood = FunctionWrapper(
            likelihood,
            likelihood_args,
            likelihood_kwargs
        )

        # Number of parameters
        if n_dim is None:
            self.n_dim = self.prior.dim
        else:
            self.n_dim = int(n_dim)

        # Effective Sample Size
        self.n_ess = int(n_ess)

        # Number of active particles
        self.n_active = int(n_active)

        # Maximum number of MCMC steps
        if max_steps is None:
            self.max_steps = 5 * self.n_dim 
        else:
            self.max_steps = int(max_steps)

        # Total number of effectively  independent samples
        self.n_total = None

        # Particle manager
        if ess_threshold is None:
            self.ess_threshold = int(4 * self.n_dim)
        else:
            self.ess_threshold = ess_threshold

        self.particles = Particles(n_active, n_dim, ess_threshold)

        # Sampling
        self.t = 0

        # Parallelism
        self.pool = pool
        if pool is None:
            self.distribute = map
        else:
            self.distribute = pool.map
        self.vectorize_likelihood = vectorize

        # Geometry
        self.u_geometry = Geometry()
        self.theta_geometry = Geometry()

        # Flow
        self.flow = Flow(self.n_dim, flow)
        self.train_config = dict(validation_split=0.5,
                                 epochs=2000,
                                 batch_size=512,
                                 patience=50,
                                 learning_rate=1e-3,
                                 annealing=False,
                                 gaussian_scale=None,
                                 laplace_scale=None,
                                 noise=None,
                                 shuffle=True,
                                 clip_grad_norm=1.0,
                                 verbose=0,
                                )
        if train_config is not None:
            for key in train_config.keys():
                self.train_config[key] = train_config[key]

        # Scaler
        self.scaler = Reparameterize(self.n_dim, bounds=self.bounds)

        # Output
        if output_dir is None:
            self.output_dir = Path("states")
        else:
            self.output_dir = output_dir
        if output_label is None:
            self.output_label = "pmc"
        else:
            self.output_label = output_label

        # Other
        self.preconditioned = precondition

        if sample is None:
            self.sample = 'pcn'
        elif sample in ['pcn']:
            self.sample = 'pcn'
        elif sample in ['rwm', 'mh']:
            self.sample = 'rwm'

        # Prior samples to draw
        if n_prior is None:
            self.n_prior = int(2 * (self.n_ess//self.n_active) * self.n_active)
        else:
            self.n_prior = int((n_prior/self.n_active) * self.n_active)
        self.prior_samples = None

        self.patience = patience

        self.logz = None
        self.logz_err = None

        self.current_particles = None
        self.warmup = True
        self.calls = 0
        
        self.progress = None
        self.pbar = None

    def run(self,
            n_total: int = 5000,
            n_evidence: int = 5000,
            progress: bool = True,
            resume_state_path: Union[str, Path] = None,
            save_every: int = None):
        r"""Run Preconditioned Monte Carlo.

        Parameters
        ----------
        n_total : int
            The total number of effectively independent samples to be
            collected (default is ``n_total=5000``).
        n_evidence : int
            The number of importance samples used to estimate the
            evidence (default is ``n_evidence=5000``). If ``n_evidence=0``,
            the evidence is not estimated using importance sampling and the
            SMC estimate is used instead. If ``preconditioned=False``, 
            the evidence is estimated using SMC and ``n_evidence`` is ignored.
        progress : bool
            If True, print progress bar (default is ``progress=True``).
        resume_state_path : ``Union[str, Path]``
            Path of state file used to resume a run. Default is ``None`` in which case
            the sampler does not load any previously saved states. An example of using
            this option to resume or continue a run is e.g. ``resume_state_path = "states/pmc_1.state"``.
        save_every : ``int`` or ``None``
            Argument which determines how often (i.e. every how many iterations) ``pocoMC`` saves
            state files to the ``output_dir`` directory. Default is ``None`` in which case no state
            files are stored during the run.
        """
        if resume_state_path is not None:
            self.load_state(resume_state_path)
            t0 = self.t
            # Initialise progress bar
            self.pbar = ProgressBar(self.progress)
            self.pbar.update_stats(dict(calls=self.particles.get("calls", -1),
                                        beta=self.particles.get("beta", -1),
                                        logZ=self.particles.get("logz", -1)))
        else:
            t0 = self.t
            # Run parameters
            self.n_total = int(n_total)
            self.progress = progress

            # Initialise progress bar
            self.pbar = ProgressBar(self.progress)

        # Initialise particles
        if self.prior_samples is None:
            self.prior_samples = self.sample_prior(self.n_prior)
            self.scaler.fit(self.prior_samples)

        if self.warmup:
            for i in range(self.n_prior//self.n_active):
                if save_every is not None:
                    if (self.t - t0) % int(save_every) == 0 and self.t != t0:
                        self.save_state(Path(self.output_dir) / f'{self.output_label}_{self.t}.state')
                # Set state parameters
                x = self.prior_samples[i*self.n_active:(i+1)*self.n_active]
                u = self.scaler.forward(x)
                logdetj = self.scaler.inverse(u)[1]
                logp = self.log_prior(x)
                logl = self._log_like(x)
                self.calls += self.n_active

                self.current_particles = dict(u=u,x=x,logl=logl,logp=logp,logdetj=logdetj,
                                    logw=-1e300 * np.ones(self.n_active), iter=self.t,
                                    calls=self.calls, steps=1, efficiency=1.0, ess=0.0, 
                                    accept=1.0, beta=0.0, logz=0.0)
                
                self.particles.update(self.current_particles)

                self.pbar.update_stats(dict(calls=self.particles.get("calls", -1), 
                                            beta=self.particles.get("beta", -1), 
                                            logZ=self.particles.get("logz", -1)))
                
                self.pbar.update_iter()

                self.t += 1
            self.warmup = False

        # Run Sequential Monte Carlo
        while self._not_termination(self.current_particles):
            if save_every is not None:
                if (self.t - t0) % int(save_every) == 0 and self.t != t0:
                    self.save_state(Path(self.output_dir) / f'{self.output_label}_{self.t}.state')

            # Choose next beta based on ESS of weights
            self.current_particles = self._reweight(self.current_particles)

            # Train Preconditioner
            self.current_particles = self._train(self.current_particles)

            # Resample particles
            self.current_particles = self._resample(self.current_particles)

            # Evolve particles using MCMC
            self.current_particles = self._mutate(self.current_particles)   

            # Save particles
            self.particles.update(self.current_particles)

        # Compute evidence
        if n_evidence > 0 and self.preconditioned:
            self._compute_evidence(int(n_evidence))
        else:
            self.logz = self.particles.compute_logz(1.0)
            self.logz_err = None

        self.pbar.close()

    def _not_termination(self, current_particles):
        """
        Check if termination criterion is satisfied.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.
        
        Returns
        -------
        termination : bool
            True if termination criterion is not satisfied.
        """
        log_weights = self.particles.compute_logw(1.0, self.ess_threshold)
        weights = np.exp(log_weights - np.max(log_weights))
        weights /= np.sum(weights)
        ess = 1.0 / np.sum(weights**2.0)

        return 1.0 - current_particles.get("beta") >= 1e-4 or ess < self.n_total

    
    def _mutate(self, current_particles):
        """
        Evolve particles using MCMC.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.
        
        Returns
        -------
        current_particles : dict
            Dictionary containing the updated particles.
        """
        state_dict = dict(
            u=current_particles.get("u").copy(),
            x=current_particles.get("x").copy(),
            logdetj=current_particles.get("logdetj").copy(),
            logp=current_particles.get("logp").copy(),
            logl=current_particles.get("logl").copy(),
            beta=current_particles.get("beta"),
        )

        function_dict = dict(
            loglike=self._log_like,
            logprior=self.log_prior,
            scaler=self.scaler,
            flow=self.flow,
            u_geometry=self.u_geometry,
            theta_geometry=self.theta_geometry,
        )

        option_dict = dict(
            nmax=self.max_steps,
            progress_bar=self.pbar,
            patience=self.patience,
        )

        if self.preconditioned and self.sample == "pcn":
            results = preconditioned_pcn(
                state_dict,
                function_dict,
                option_dict
                )
        elif self.preconditioned and self.sample == "rwm":
            results = preconditioned_rwm(
                state_dict,
                function_dict,
                option_dict
                )
        elif not self.preconditioned and self.sample == "pcn":
            results = pcn(
                state_dict,
                function_dict,
                option_dict
                )
        elif not self.preconditioned and self.sample == "rwm":
            results = rwm(
                state_dict,
                function_dict,
                option_dict
                )

        current_particles["u"] = results.get('u').copy()
        current_particles["x"] = results.get('x').copy()
        current_particles["logdetj"] = results.get('logdetj').copy()
        current_particles["logl"] = results.get('logl').copy()
        current_particles["logp"] = results.get('logp').copy()
        current_particles["efficiency"] = results.get('efficiency') / (2.38 / self.n_dim ** 0.5)
        current_particles["steps"] = results.get('steps')
        current_particles["accept"] = results.get('accept')
        current_particles["calls"] = current_particles.get("calls") + results.get('calls')

        return current_particles


    def _train(self, current_particles):
        """
        Train normalizing flow.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.
        
        Returns
        -------
        current_particles : dict
            Dictionary containing the updated particles.
        """
        u = current_particles.get("u")
        w = current_particles.get("weights")

        self.u_geometry.fit(u, weights=w)

        if self.preconditioned:

            self.flow.fit(numpy_to_torch(u),
                          weights=numpy_to_torch(w),
                          validation_split=self.train_config["validation_split"],
                          epochs=self.train_config["epochs"],
                          batch_size=int(np.minimum(len(u)//2, self.train_config["batch_size"])),
                          gaussian_scale=self.train_config["gaussian_scale"],
                          laplace_scale=self.train_config["laplace_scale"],
                          patience=self.train_config["patience"],
                          learning_rate=self.train_config["learning_rate"],
                          annealing=self.train_config["annealing"],
                          noise=self.train_config["noise"],
                          shuffle=self.train_config["shuffle"],
                          clip_grad_norm=self.train_config["clip_grad_norm"],
                          verbose=self.train_config["verbose"],
                          )
            
            theta = flow_numpy_wrapper(self.flow).forward(u)[0]
            self.theta_geometry.fit(theta, weights=w)
        else:
            pass



        return current_particles

    def _resample(self, current_particles):
        """
        Resample particles.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.

        Returns
        -------
        current_particles : dict
            Dictionary containing the updated particles.
        """
        u = current_particles.get("u")
        x = current_particles.get("x")
        logdetj = current_particles.get("logdetj")
        logl = current_particles.get("logl")
        logp = current_particles.get("logp")
        weights = current_particles.get("weights")

        idx_resampled = systematic_resample(self.n_active, weights=weights)

        current_particles["u"] = u[idx_resampled]
        current_particles["x"] = x[idx_resampled]
        current_particles["logdetj"] = logdetj[idx_resampled]
        current_particles["logl"] = logl[idx_resampled]
        current_particles["logp"] = logp[idx_resampled]
    
        return current_particles

    def _reweight(self, current_particles):
        """
        Reweight particles.

        Parameters
        ----------
        current_particles : dict
            Dictionary containing the current particles.

        Returns
        -------
        current_particles : dict
            Dictionary containing the updated particles.
        """
        # Update iteration index
        self.t += 1
        self.pbar.update_iter()

        beta_prev = self.particles.get("beta", index=-1)
        beta_max = 1.0
        beta_min = np.copy(beta_prev)

        def get_weights_and_ess(beta):
            logw = self.particles.compute_logw(beta)
            weights = np.exp(logw - np.max(logw))
            weights /= np.sum(weights)
            ess_est = 1.0 / np.sum(weights**2.0)
            return weights, ess_est

        weights_prev, ess_est_prev = get_weights_and_ess(beta_prev)
        weights_max, ess_est_max = get_weights_and_ess(beta_max)


        if ess_est_prev <= self.n_ess:
            beta = beta_prev
            weights = weights_prev
            logz = self.particles.get("logz", index=-1)
            self.pbar.update_stats(dict(beta=beta, ESS=ess_est_prev, logZ=logz))
        elif ess_est_max >= self.n_ess:
            beta = beta_max 
            weights = weights_max
            logz = self.particles.compute_logz(beta)
            self.pbar.update_stats(dict(beta=beta, ESS=ess_est_max, logZ=logz))
        else:
            while True:
                beta = (beta_max + beta_min) * 0.5

                weights, ess_est = get_weights_and_ess(beta)

                if np.abs(ess_est - self.n_ess) < 0.01 * self.n_ess or beta == 1.0:
                    logz = self.particles.compute_logz(beta)
                    self.pbar.update_stats(dict(beta=beta, ESS=ess_est, logZ=logz))
                    break
                elif ess_est < self.n_ess:
                    beta_max = beta
                else:
                    beta_min = beta

        logw = self.particles.compute_logw(beta, self.ess_threshold)
        weights = np.exp(logw - np.max(logw))
        weights /= np.sum(weights)

        idx, weights = trim_weights(np.arange(len(weights)), weights, ess=0.99, bins=1000)
        current_particles["u"] = self.particles.get("u", index=None, flat=True)[idx]
        current_particles["x"] = self.particles.get("x", index=None, flat=True)[idx]
        current_particles["logdetj"] = self.particles.get("logdetj", index=None, flat=True)[idx]
        current_particles["logl"] = self.particles.get("logl", index=None, flat=True)[idx]
        current_particles["logp"] = self.particles.get("logp", index=None, flat=True)[idx]
        current_particles["logz"] = logz
        current_particles["beta"] = beta
        current_particles["weights"] = weights

        return current_particles

    def _log_like(self, x):
        """
        Compute log likelihood.

        Parameters
        ----------
        x : array_like
            Array of parameter values.
        
        Returns
        -------
        logl : float
            Log likelihood.
        """
        if self.vectorize_likelihood:
            return self.log_likelihood(x)
        elif self.pool is not None:
            return np.array(list(self.distribute(self.log_likelihood, x)))
        else:
            return np.array(list(map(self.log_likelihood, x)))
        
    def evidence(self):
        """
        Return the log evidence estimate and error.
        """
        return self.logz, self.logz_err
        
    def _compute_evidence(self, n=5_000):
        """
        Estimate the evidence using importance sampling.

        Parameters
        ----------
        n : int
            Number of importance samples (default is ``n=5_000``).
        
        Returns
        -------
        logz : float
            Estimate of the log evidence.
        dlogz : float
            Estimate of the error on the log evidence.
        """
        with torch.no_grad():
            theta_q, logq = self.flow.sample(n)
            theta_q = torch_to_numpy(theta_q)
            logq = torch_to_numpy(logq)

        x_q, logdetj = self.scaler.inverse(theta_q)
        logl = self._log_like(x_q)
        logp = self.log_prior(x_q)

        logw = logl + logp + logdetj - logq 
        logz = np.logaddexp.reduce(logw) - np.log(len(logw))

        dlogz = np.std([np.logaddexp.reduce(logw[np.random.choice(len(logw), len(logw))]) - np.log(len(logw)) for _ in range(np.maximum(n,1000))])

        self.logz = logz
        self.logz_err = dlogz
        return logz, dlogz

    def __getstate__(self):
        """
        Get state information for pickling.
        """
        state = self.__dict__.copy()

        try:
            # remove random module
            # del state['rstate']

            # deal with pool
            if state['pool'] is not None:
                del state['pool']  # remove pool
                del state['distribute']  # remove `pool.map` function hook
        except:  # TODO use specific exception type
            pass

        return state

    def posterior(self, resample=False, trim_importance_weights=True, return_logw=False, ess_trim=0.99, bins_trim=1_000):
        """
        Return posterior samples.

        Parameters
        ----------
        resample : bool
            If True, resample particles (default is ``resample=False``).
        trim_importance_weights : bool
            If True, trim importance weights (default is ``trim_importance_weights=True``).
        return_logw : bool
            If True, return log importance weights (default is ``return_logw=False``).
        ess_trim : float
            Effective sample size threshold for trimming (default is ``ess_trim=0.99``).
        bins_trim : int
            Number of bins for trimming (default is ``bins_trim=1_000``).

        Returns
        -------
        samples : ``np.ndarray``
            Samples from the posterior.
        weights : ``np.ndarray``
            Importance weights.
        logl : ``np.ndarray``
            Log likelihoods.
        logp : ``np.ndarray``
            Log priors.
        """
        samples = self.particles.get("x", flat=True)
        logl = self.particles.get("logl", flat=True)
        logp = self.particles.get("logp", flat=True)
        logw = self.particles.compute_logw(1.0, self.ess_threshold)
        weights = np.exp(logw)

        if trim_importance_weights:
            idx, weights = trim_weights(np.arange(len(samples)), weights, ess=ess_trim, bins=bins_trim)
            samples = samples[idx]
            logl = logl[idx]
            logp = logp[idx]
            logw = logw[idx]

        if resample:
            idx_resampled = systematic_resample(len(weights), weights=weights)
            return samples[idx_resampled], logl[idx_resampled], logp[idx_resampled]
            
        else:
            if return_logw:
                return samples, logw, logl, logp
            else:
                return samples, weights, logl, logp

    @property
    def results(self):
        """
        Return results.

        Returns
        -------
        results : dict
            Dictionary containing the results.
        """
        return self.particles.compute_results()

    def save_state(self, path: Union[str, Path]):
        """Save current state of sampler to file.

        Parameters
        ----------
        path : ``Union[str, Path]``
            Path to save state.
        """
        print(f'Saving PMC state to {path}')
        Path(path).parent.mkdir(exist_ok=True)
        with open(path, 'wb') as f:
            state = self.__dict__.copy()
            del state['pbar']  # Cannot be pickled
            try:
                # deal with pool
                if state['pool'] is not None:
                    del state['pool']  # remove pool
                    del state['distribute']  # remove `pool.map` function hook
            except BaseException as e:
                print(e)
            dill.dump(file=f, obj=state)

    def load_state(self, path: Union[str, Path]):
        """Load state of sampler from file.

        Parameters
        ----------
        path : ``Union[str, Path]``
            Path from which to load state.
        """
        with open(path, 'rb') as f:
            state = dill.load(file=f)
        self.__dict__ = {**self.__dict__, **state}

